package at.ac.uibk.gitsearch.service;

import at.ac.uibk.gitsearch.domain.ChildInfo;
import at.ac.uibk.gitsearch.repository.gitlab.GitLabRepository;
import at.ac.uibk.gitsearch.repository.search.MetaDataRepository;
import at.ac.uibk.gitsearch.security.AuthoritiesConstants.AuthoritiesConstantEnum;
import at.ac.uibk.gitsearch.security.jwt.TokenProvider;
import at.ac.uibk.gitsearch.security.jwt.TokenProvider.GitLabAccessInfo;
import java.io.File;
import java.io.IOException;
import java.net.URI;
import java.net.URISyntaxException;
import java.text.ParseException;
import java.util.Arrays;
import java.util.HashSet;
import java.util.List;
import java.util.Optional;
import java.util.stream.Collectors;
import java.util.stream.Stream;
import java.util.zip.ZipInputStream;
import org.codeability.sharing.plugins.api.search.SearchInputDTO;
import org.codeability.sharing.plugins.api.search.SearchInputMetadataDTO;
import org.codeability.sharing.plugins.api.search.SearchResultDTO;
import org.codeability.sharing.plugins.api.search.SearchResultsDTO;
import org.codeability.sharing.plugins.api.search.util.ExerciseId;
import org.gitlab4j.api.GitLabApi;
import org.gitlab4j.api.GitLabApiException;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.security.core.userdetails.User;
import org.springframework.stereotype.Service;
import org.springframework.web.servlet.support.ServletUriComponentsBuilder;

/**
 * Service for exercise/course search results
 * <p>
 */
@Service
@SuppressWarnings("PMD")
public class SearchService {

    private final Logger log = LoggerFactory.getLogger(SearchService.class);

    private static final String ZIP = "zip";

    @Autowired
    protected MetaDataRepository metaDataRepository;

    @Autowired
    protected TokenProvider tokenProvider;

    @Autowired
    protected GitLabRepository gitLabRepository;

    @Autowired
    protected GitlabService gitlabService;

    @Autowired
    protected PluginManagementService pluginManagementService;

    @Autowired
    protected ShoppingBasketService shoppingBasketService;

    @Autowired
    private FileService fileService;

    @Autowired
    private UserService userService;

    /**
     * returns the result page
     *
     * @param searchInput the query definition
     */
    public SearchResultsDTO searchResultPage(SearchInputDTO searchInput) throws IOException {
        final Optional<User> principal = tokenProvider.getCurrentPrincipal();
        log.debug("Searchrequest for {} ", searchInput);

        final SearchResultsDTO pageDetails = metaDataRepository.pageDetailsWithJavaApi(searchInput, principal);

        pageDetails
            .getSearchResult()
            .stream()
            .forEach(hit ->
                pluginManagementService
                    .getRegisteredPluginConfigs()
                    .stream()
                    .forEach(config ->
                        config
                            .getActions()
                            .values()
                            .stream()
                            .filter(action -> action.isApplicable(hit))
                            .forEach(action -> hit.getSupportedActions().add(action.getPluginActionInfo()))
                    )
            );
        pageDetails.getSearchResult().stream().forEach(this::fixImageURL);
        pageDetails.getSearchResult().forEach(this::filterArtemisActionForUnauthorized);

        return pageDetails;
    }

    public SearchResultsDTO searchResultsForLoggedInUser() throws IOException {
        at.ac.uibk.gitsearch.domain.User user = userService.getUserWithAuthorities().orElse(null);
        if (user == null) {
            return new SearchResultsDTO();
        }
        return searchResultForAuthorEmail(user.getEmail(), 1000);
    }

    public SearchResultsDTO searchResultForAuthorEmail(String searchInput, int length) throws IOException {
        log.debug("Searchrequest for {} ", searchInput);

        final SearchResultsDTO pageDetails = metaDataRepository.searchByEmail(searchInput, length, tokenProvider.getCurrentPrincipal());

        pageDetails
            .getSearchResult()
            .stream()
            .forEach(hit ->
                pluginManagementService
                    .getRegisteredPluginConfigs()
                    .stream()
                    .forEach(config ->
                        config
                            .getActions()
                            .values()
                            .stream()
                            .filter(action -> action.isApplicable(hit))
                            .forEach(action -> hit.getSupportedActions().add(action.getPluginActionInfo()))
                    )
            );
        pageDetails.getSearchResult().stream().forEach(this::fixImageURL);

        return pageDetails;
    }

    // Returns all resources indexed by elastic search
    public List<SearchResultDTO> getAllResources() throws IOException {
        return metaDataRepository.getAllResources();
    }

    public enum ExtractionDepth {
        JUST_PROJECT,
        WITH_CHILDREN,
        WITH_DESCENDANTS,
    }

    /**
     * returns the result page
     *
     * @param page        the index of the first record to be returned
     * @param length      the number of records on page
     */
    public SearchResultsDTO watchListResultPage(Stream<String> exerciseIds, int page, int length) {
        final SearchResultsDTO pageDetails = metaDataRepository.getExercisesById(
            exerciseIds,
            tokenProvider.getCurrentPrincipal(),
            page,
            length
        );

        pageDetails
            .getSearchResult()
            .stream()
            .forEach(hit ->
                pluginManagementService
                    .getRegisteredPluginConfigs()
                    .stream()
                    .forEach(config ->
                        config
                            .getActions()
                            .values()
                            .stream()
                            .filter(action -> action.isApplicable(hit))
                            .forEach(action -> hit.getSupportedActions().add(action.getPluginActionInfo()))
                    )
            );
        pageDetails.getSearchResult().stream().forEach(this::fixImageURL);

        return pageDetails;
    }

    private void fixImageURL(SearchResultDTO metaData) {
        String image = metaData.getMetadata().getImage();

        try {
            if (image != null) {
                URI url = new URI(image);
                if (url.isAbsolute()) {
                    return;
                }
                String httpUrlToRepo = metaData.getProject().getUrl();
                String commitId = metaData.getFile().getCommit_id();
                String baseRepoURL = httpUrlToRepo + "/-/raw/" + commitId + "/";
                final URI resolvedImageUrl = new URI(baseRepoURL).resolve(url);
                metaData.getMetadata().setImage(resolvedImageUrl.toASCIIString());
                return;
            }

            final String baseUrl = ServletUriComponentsBuilder.fromCurrentContextPath().build().toUriString();
            final boolean isJava =
                metaData.getMetadata().getProgrammingLanguage() != null &&
                Arrays.stream(metaData.getMetadata().getProgrammingLanguage()).anyMatch(s -> s.startsWith("JAVA")) &&
                metaData.hashCode() % 2 == 0; // nur um ein wenig Abwechslung zu bekommen :-)
            if (isJava) {
                metaData.getMetadata().setImage(baseUrl + "/content/img/java.png");
                return;
            }
            final boolean isPython =
                metaData.getMetadata().getProgrammingLanguage() != null &&
                Arrays.stream(metaData.getMetadata().getProgrammingLanguage()).anyMatch(s -> s.startsWith("python"));
            if (isPython) {
                metaData.getMetadata().setImage(baseUrl + "/content/img/python.png");
                return;
            }
            metaData.getMetadata().setImage(baseUrl + "/content/images/Logo_codeAbility_4c_300dpi_RGB3.gif");
        } catch (URISyntaxException e) {
            log.warn("Cannot parse image url {} for {}", metaData.getMetadata().getImage(), metaData.getProject().getUrl());
        }
    }

    public File exportExercise(String exerciseId) throws ParseException {
        return exportExercise(ExerciseId.fromString(exerciseId), ExtractionDepth.JUST_PROJECT);
    }

    public File exportExercise(ExerciseId exerciseId, ExtractionDepth recursive) {
        try (
            ZipInputStream inputFile = fileService.getZippedExercise(
                exerciseId,
                tokenProvider.getCurrentPrincipal(),
                recursive,
                new HashSet<>()
            );
        ) {
            if (inputFile == null) {
                return null;
            }
            File file = File.createTempFile("exercise" + exerciseId.getProjectId(), ".zip");
            return fileService.copyInputStreamToFile(inputFile, file);
        } catch (IOException e) {
            log.error("Error while exporting exercise", e);
        }
        return null;
    }

    public Boolean hasAccessToExerciseId(String exerciseId) {
        final Optional<User> currentPrincipalO = tokenProvider.getCurrentPrincipal();
        if (currentPrincipalO.isEmpty()) {
            log.warn("Cannot find a principal for for exercise {}", exerciseId);
            return false;
        } else if (currentPrincipalO.get().getAuthorities().contains(AuthoritiesConstantEnum.ADMIN.getGrantedAuthority())) {
            log.info("Admin can access everything : {}", exerciseId);
            return true; // ADMIN is always allowed
        }
        SearchResultDTO result = metaDataRepository.getExerciseById(exerciseId, currentPrincipalO);
        if (result == null) {
            log.warn("Cannot find exercise for : {}", exerciseId);
            return false;
        }
        return true;
    }

    public Optional<SearchResultDTO> findExerciseById(ExerciseId exerciseId) {
        try {
            SearchResultDTO result = metaDataRepository.getExerciseById(exerciseId, tokenProvider.getCurrentPrincipal());
            return Optional.ofNullable(result);
        } catch (javax.ws.rs.NotFoundException e) {
            log.error("exercise with id {} not found?", exerciseId, e);
            return Optional.empty();
        }
    }

    public List<ChildInfo> findChildrenByExerciseId(ExerciseId exerciseId) {
        return metaDataRepository.getChildInfos(exerciseId, tokenProvider.getCurrentPrincipal());
    }

    /**
     * Method used to filter out the Artemis export action for
     * non-authorized users having only partial access to the data.
     *
     * See Metadata entry 'publicVisibility.except'
     * @param result SearchResultDTO to filter the actions from
     * @return the updated SearchResultDTO with filtered actions
     */
    private SearchResultDTO filterArtemisActionForUnauthorized(SearchResultDTO result) {
        if (result.getMetadata().getPublicVisibility() == null || result.getMetadata().getPublicVisibility().getExcept() == null) {
            // results queried without 'publicVisibility' are already authorized
            return result;
        }
        try {
            final Optional<GitLabAccessInfo> gitLabAccessInfo = tokenProvider.getGitLabAccessInfo();
            if (gitLabAccessInfo.isEmpty()) {
                log.warn("User unauthorized to export: {}", result.getProject().getUrl());
                result.setSupportedActions(
                    result
                        .getSupportedActions()
                        .stream()
                        .filter(action -> !"Artemis Sharing Connector".equals(action.getPlugin()))
                        .collect(Collectors.toList())
                );
                return result;
            }

            GitLabApi gitLabApi = gitLabRepository.getGitLabApi(gitLabAccessInfo);
            gitLabApi.getRepositoryApi().getRepositoryArchive(Long.toString(result.getProject().getProject_id()), "HEAD", ZIP);
        } catch (GitLabApiException e) {
            log.warn(e.getMessage());
            // We expect permissions missing
            result.setSupportedActions(
                result
                    .getSupportedActions()
                    .stream()
                    .filter(action -> !"Artemis Sharing Connector".equals(action.getPlugin()))
                    .collect(Collectors.toList())
            );
        }
        return result;
    }

    /**
     * Method to stream all exercises.
     *
     * @return the search result page DTO
     * @throws IOException
     */
    public Stream<SearchResultDTO> getAllResultsAsStream() throws IOException {
        return metaDataRepository.getAllResultsAsStream();
    }

    /**
     * Method to search for all public exercises.
     * Searches all public exercises with an empty search query and without a
     * specified user.
     *
     * @return the search result page DTO
     * @throws IOException
     */
    public Stream<SearchResultDTO> getAllPublicResultsAsStream() throws IOException {
        log.debug("Search request to find all public exercises");

        final SearchInputMetadataDTO emptySearchMetadata = new SearchInputMetadataDTO(null, null, null, null, null, null);
        SearchInputDTO emptySearchQuery = new SearchInputDTO(null, emptySearchMetadata, null, null, null, 0);

        return metaDataRepository.streamResults(emptySearchQuery, Optional.empty(), false).getSearchStream();
    }
}
