package at.ac.uibk.gitsearch.web.rest;

import at.ac.uibk.gitsearch.domain.ChatMessage;
import at.ac.uibk.gitsearch.domain.Review;
import at.ac.uibk.gitsearch.domain.User;
import at.ac.uibk.gitsearch.service.ChatMessageService;
import at.ac.uibk.gitsearch.service.ReviewService;
import at.ac.uibk.gitsearch.service.UserService;
import at.ac.uibk.gitsearch.service.dto.ChatMessageDTO;
import java.util.ArrayList;
import java.util.List;
import org.springframework.http.ResponseEntity;
import org.springframework.security.access.AccessDeniedException;
import org.springframework.security.access.prepost.PreAuthorize;
import org.springframework.web.bind.annotation.DeleteMapping;
import org.springframework.web.bind.annotation.GetMapping;
import org.springframework.web.bind.annotation.PathVariable;
import org.springframework.web.bind.annotation.PostMapping;
import org.springframework.web.bind.annotation.RequestBody;
import org.springframework.web.bind.annotation.RequestMapping;
import org.springframework.web.bind.annotation.RestController;

@RestController
@RequestMapping("/api/chatmessages")
public class ChatMessageResource {

    private final ChatMessageService chatMessageService;

    private final UserService userService;

    private final ReviewService reviewService;

    public ChatMessageResource(ChatMessageService chatMessageService, UserService userService, ReviewService reviewService) {
        this.chatMessageService = chatMessageService;
        this.userService = userService;
        this.reviewService = reviewService;
    }

    @GetMapping
    @PreAuthorize("hasAnyRole('ADMIN')")
    public List<ChatMessage> getAllChatMessages() {
        return chatMessageService.getAllChatMessages();
    }

    @GetMapping("/{id}")
    @PreAuthorize("hasAnyRole('ADMIN')")
    public ResponseEntity<ChatMessage> getChatMessageById(@PathVariable Long id) {
        return chatMessageService.getChatMessageById(id).map(ResponseEntity::ok).orElse(ResponseEntity.notFound().build());
    }

    @PostMapping
    @PreAuthorize("hasAnyRole('USER')")
    public ChatMessageDTO saveChatMessage(@RequestBody ChatMessageDTO chatMessage) {
        return chatMessageService.toDTO(chatMessageService.saveChatMessage(chatMessage));
    }

    @DeleteMapping("/{id}")
    @PreAuthorize("hasAnyRole('ADMIN')")
    public ResponseEntity<Void> deleteChatMessage(@PathVariable Long id) {
        chatMessageService.deleteChatMessage(id);
        return ResponseEntity.ok().build();
    }

    @GetMapping("/review/{reviewId}")
    @PreAuthorize("hasAnyRole('USER')")
    public List<ChatMessageDTO> findByReview(@PathVariable Long reviewId) {
        Review review = reviewService.getReviewById(reviewId);
        User user = userService.getUserWithAuthorities().get();
        if (!chatMessageService.hasAccessToReview(user, review)) {
            throw new AccessDeniedException("You do not have access to this review");
        }

        List<ChatMessage> chatMessages = chatMessageService.findByReview(review);
        List<ChatMessageDTO> chatMessageDTOS = new ArrayList<>();
        for (ChatMessage chatMessage : chatMessages) {
            chatMessageDTOS.add(chatMessageService.toDTO(chatMessage));
        }

        return chatMessageService.replaceSenderNamesWithReviewerNames(chatMessageDTOS);
    }

    @GetMapping("/review/{reviewId}/ordered")
    @PreAuthorize("hasAnyRole('USER')")
    public List<ChatMessageDTO> findByReviewOrderByTimestampAsc(@PathVariable Long reviewId) {
        Review review = reviewService.getReviewById(reviewId);
        User user = userService.getUserWithAuthorities().get();
        if (!chatMessageService.hasAccessToReview(user, review)) {
            throw new AccessDeniedException("You do not have access to this review");
        }
        List<ChatMessage> chatMessages = chatMessageService.findByReviewOrderByTimestampAsc(review);
        List<ChatMessageDTO> chatMessageDTOS = new ArrayList<>();
        for (ChatMessage chatMessage : chatMessages) {
            chatMessageDTOS.add(chatMessageService.toDTO(chatMessage));
        }

        return chatMessageService.replaceSenderNamesWithReviewerNames(chatMessageDTOS);
    }
}
