package at.ac.uibk.gitsearch.web.rest;

import at.ac.uibk.gitsearch.repository.jpa.ReviewRatingRepository;
import at.ac.uibk.gitsearch.service.ReviewRatingQueryService;
import at.ac.uibk.gitsearch.service.ReviewRatingService;
import at.ac.uibk.gitsearch.service.criteria.ReviewRatingCriteria;
import at.ac.uibk.gitsearch.service.dto.ReviewRatingDTO;
import at.ac.uibk.gitsearch.web.rest.errors.BadRequestAlertException;
import java.net.URI;
import java.net.URISyntaxException;
import java.util.List;
import java.util.Objects;
import java.util.Optional;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.http.ResponseEntity;
import org.springframework.web.bind.annotation.DeleteMapping;
import org.springframework.web.bind.annotation.GetMapping;
import org.springframework.web.bind.annotation.PatchMapping;
import org.springframework.web.bind.annotation.PathVariable;
import org.springframework.web.bind.annotation.PostMapping;
import org.springframework.web.bind.annotation.PutMapping;
import org.springframework.web.bind.annotation.RequestBody;
import org.springframework.web.bind.annotation.RequestMapping;
import org.springframework.web.bind.annotation.RestController;
import tech.jhipster.web.util.HeaderUtil;
import tech.jhipster.web.util.ResponseUtil;

/**
 * REST controller for managing {@link at.ac.uibk.gitsearch.domain.ReviewRating}.
 */
@RestController
@RequestMapping("/api")
@SuppressWarnings({ "PMD.AvoidDuplicateLiterals" })
public class ReviewRatingResource {

    private final Logger log = LoggerFactory.getLogger(ReviewRatingResource.class);

    private static final String ENTITY_NAME = "reviewRating";

    @Value("${jhipster.clientApp.name}")
    private String applicationName;

    private final ReviewRatingService reviewRatingService;

    private final ReviewRatingRepository reviewRatingRepository;

    private final ReviewRatingQueryService reviewRatingQueryService;

    public ReviewRatingResource(
        ReviewRatingService reviewRatingService,
        ReviewRatingRepository reviewRatingRepository,
        ReviewRatingQueryService reviewRatingQueryService
    ) {
        this.reviewRatingService = reviewRatingService;
        this.reviewRatingRepository = reviewRatingRepository;
        this.reviewRatingQueryService = reviewRatingQueryService;
    }

    /**
     * {@code POST  /review-ratings} : Create a new reviewRating.
     *
     * @param reviewRatingDTO the reviewRatingDTO to create.
     * @return the {@link ResponseEntity} with status {@code 201 (Created)} and with body the new reviewRatingDTO, or with status {@code 400 (Bad Request)} if the reviewRating has already an ID.
     * @throws URISyntaxException if the Location URI syntax is incorrect.
     */
    @PostMapping("/review-ratings")
    public ResponseEntity<ReviewRatingDTO> createReviewRating(@RequestBody ReviewRatingDTO reviewRatingDTO) throws URISyntaxException {
        log.debug("REST request to save ReviewRating : {}", reviewRatingDTO);
        if (reviewRatingDTO.getId() != null) {
            throw new BadRequestAlertException("A new reviewRating cannot already have an ID", ENTITY_NAME, "idexists");
        }
        ReviewRatingDTO result = reviewRatingService.save(reviewRatingDTO);
        return ResponseEntity
            .created(new URI("/api/review-ratings/" + result.getId()))
            .headers(HeaderUtil.createEntityCreationAlert(applicationName, true, ENTITY_NAME, result.getId().toString()))
            .body(result);
    }

    /**
     * {@code PUT  /review-ratings/:id} : Updates an existing reviewRating.
     *
     * @param id the id of the reviewRatingDTO to save.
     * @param reviewRatingDTO the reviewRatingDTO to update.
     * @return the {@link ResponseEntity} with status {@code 200 (OK)} and with body the updated reviewRatingDTO,
     * or with status {@code 400 (Bad Request)} if the reviewRatingDTO is not valid,
     * or with status {@code 500 (Internal Server Error)} if the reviewRatingDTO couldn't be updated.
     * @throws URISyntaxException if the Location URI syntax is incorrect.
     */
    @PutMapping("/review-ratings/{id}")
    public ResponseEntity<ReviewRatingDTO> updateReviewRating(
        @PathVariable(value = "id", required = false) final Long id,
        @RequestBody ReviewRatingDTO reviewRatingDTO
    ) throws URISyntaxException {
        log.debug("REST request to update ReviewRating : {}, {}", id, reviewRatingDTO);
        if (reviewRatingDTO.getId() == null) {
            throw new BadRequestAlertException("Invalid id", ENTITY_NAME, "idnull");
        }
        if (!Objects.equals(id, reviewRatingDTO.getId())) {
            throw new BadRequestAlertException("Invalid ID", ENTITY_NAME, "idinvalid");
        }

        if (!reviewRatingRepository.existsById(id)) {
            throw new BadRequestAlertException("Entity not found", ENTITY_NAME, "idnotfound");
        }

        ReviewRatingDTO result = reviewRatingService.save(reviewRatingDTO);
        return ResponseEntity
            .ok()
            .headers(HeaderUtil.createEntityUpdateAlert(applicationName, true, ENTITY_NAME, reviewRatingDTO.getId().toString()))
            .body(result);
    }

    /**
     * {@code PATCH  /review-ratings/:id} : Partial updates given fields of an existing reviewRating, field will ignore if it is null
     *
     * @param id the id of the reviewRatingDTO to save.
     * @param reviewRatingDTO the reviewRatingDTO to update.
     * @return the {@link ResponseEntity} with status {@code 200 (OK)} and with body the updated reviewRatingDTO,
     * or with status {@code 400 (Bad Request)} if the reviewRatingDTO is not valid,
     * or with status {@code 404 (Not Found)} if the reviewRatingDTO is not found,
     * or with status {@code 500 (Internal Server Error)} if the reviewRatingDTO couldn't be updated.
     * @throws URISyntaxException if the Location URI syntax is incorrect.
     */
    @PatchMapping(value = "/review-ratings/{id}", consumes = { "application/json", "application/merge-patch+json" })
    public ResponseEntity<ReviewRatingDTO> partialUpdateReviewRating(
        @PathVariable(value = "id", required = false) final Long id,
        @RequestBody ReviewRatingDTO reviewRatingDTO
    ) throws URISyntaxException {
        log.debug("REST request to partial update ReviewRating partially : {}, {}", id, reviewRatingDTO);
        if (reviewRatingDTO.getId() == null) {
            throw new BadRequestAlertException("Invalid id", ENTITY_NAME, "idnull");
        }
        if (!Objects.equals(id, reviewRatingDTO.getId())) {
            throw new BadRequestAlertException("Invalid ID", ENTITY_NAME, "idinvalid");
        }

        if (!reviewRatingRepository.existsById(id)) {
            throw new BadRequestAlertException("Entity not found", ENTITY_NAME, "idnotfound");
        }

        Optional<ReviewRatingDTO> result = reviewRatingService.partialUpdate(reviewRatingDTO);

        return ResponseUtil.wrapOrNotFound(
            result,
            HeaderUtil.createEntityUpdateAlert(applicationName, true, ENTITY_NAME, reviewRatingDTO.getId().toString())
        );
    }

    /**
     * {@code GET  /review-ratings} : get all the reviewRatings.
     *
     * @param criteria the criteria which the requested entities should match.
     * @return the {@link ResponseEntity} with status {@code 200 (OK)} and the list of reviewRatings in body.
     */
    @GetMapping("/review-ratings")
    public ResponseEntity<List<ReviewRatingDTO>> getAllReviewRatings(ReviewRatingCriteria criteria) {
        log.debug("REST request to get ReviewRatings by criteria: {}", criteria);
        List<ReviewRatingDTO> entityList = reviewRatingQueryService.findByCriteria(criteria);
        return ResponseEntity.ok().body(entityList);
    }

    /**
     * {@code GET  /review-ratings/count} : count all the reviewRatings.
     *
     * @param criteria the criteria which the requested entities should match.
     * @return the {@link ResponseEntity} with status {@code 200 (OK)} and the count in body.
     */
    @GetMapping("/review-ratings/count")
    public ResponseEntity<Long> countReviewRatings(ReviewRatingCriteria criteria) {
        log.debug("REST request to count ReviewRatings by criteria: {}", criteria);
        return ResponseEntity.ok().body(reviewRatingQueryService.countByCriteria(criteria));
    }

    /**
     * {@code GET  /review-ratings/:id} : get the "id" reviewRating.
     *
     * @param id the id of the reviewRatingDTO to retrieve.
     * @return the {@link ResponseEntity} with status {@code 200 (OK)} and with body the reviewRatingDTO, or with status {@code 404 (Not Found)}.
     */
    @GetMapping("/review-ratings/{id}")
    public ResponseEntity<ReviewRatingDTO> getReviewRating(@PathVariable Long id) {
        log.debug("REST request to get ReviewRating : {}", id);
        Optional<ReviewRatingDTO> reviewRatingDTO = reviewRatingService.findOne(id);
        return ResponseUtil.wrapOrNotFound(reviewRatingDTO);
    }

    /**
     * {@code DELETE  /review-ratings/:id} : delete the "id" reviewRating.
     *
     * @param id the id of the reviewRatingDTO to delete.
     * @return the {@link ResponseEntity} with status {@code 204 (NO_CONTENT)}.
     */
    @DeleteMapping("/review-ratings/{id}")
    public ResponseEntity<Void> deleteReviewRating(@PathVariable Long id) {
        log.debug("REST request to delete ReviewRating : {}", id);
        reviewRatingService.delete(id);
        return ResponseEntity
            .noContent()
            .headers(HeaderUtil.createEntityDeletionAlert(applicationName, true, ENTITY_NAME, id.toString()))
            .build();
    }
}
