import { Component, OnInit, OnDestroy } from '@angular/core';
import { HttpHeaders, HttpResponse } from '@angular/common/http';
import { ActivatedRoute } from '@angular/router';
import { Subscription } from 'rxjs';
import { EventManager } from 'app/core/util/event-manager.service';
import { NgbModal } from '@ng-bootstrap/ng-bootstrap';

import { IWatchListEntry } from 'app/shared/model/watch-list-entry.model';

import { ITEMS_PER_PAGE } from 'app/config/pagination.constants';
import { ParseLinks } from 'app/core/util/parse-links.service';
import { WatchListEntryService } from './watch-list-entry.service';
import { WatchListEntryDeleteDialogComponent } from './watch-list-entry-delete-dialog.component';

@Component({
  selector: 'jhi-watch-list-entry',
  templateUrl: './watch-list-entry.component.html',
})
export class WatchListEntryComponent implements OnInit, OnDestroy {
  watchListEntries: IWatchListEntry[];
  eventSubscriber?: Subscription;
  itemsPerPage: number;
  links: any;
  page: number;
  predicate: string;
  ascending: boolean;
  currentSearch: string;

  constructor(
    protected watchListEntryService: WatchListEntryService,
    protected eventManager: EventManager,
    protected modalService: NgbModal,
    protected parseLinks: ParseLinks,
    protected activatedRoute: ActivatedRoute
  ) {
    this.watchListEntries = [];
    this.itemsPerPage = ITEMS_PER_PAGE;
    this.page = 0;
    this.links = {
      last: 0,
    };
    this.predicate = 'id';
    this.ascending = true;
    this.currentSearch =
      this.activatedRoute.snapshot && this.activatedRoute.snapshot.queryParams['search']
        ? this.activatedRoute.snapshot.queryParams['search']
        : '';
  }

  loadAll(): void {
    if (this.currentSearch) {
      this.watchListEntryService
        .search({
          query: this.currentSearch,
          //        TODO
          //        page: this.page,
          //        size: this.itemsPerPage,
          //        sort: this.sort(),
        })
        .subscribe((res: HttpResponse<IWatchListEntry[]>) => this.paginateWatchListEntries(res.body, res.headers));
      return;
    }

    this.watchListEntryService
      .query({
        page: this.page,
        size: this.itemsPerPage,
        sort: this.sort(),
      })
      .subscribe((res: HttpResponse<IWatchListEntry[]>) => this.paginateWatchListEntries(res.body, res.headers));
  }

  reset(): void {
    this.page = 0;
    this.watchListEntries = [];
    this.loadAll();
  }

  loadPage(page: number): void {
    this.page = page;
    this.loadAll();
  }

  search(query: string): void {
    this.watchListEntries = [];
    this.links = {
      last: 0,
    };
    this.page = 0;
    if (query) {
      this.predicate = '_score';
      this.ascending = false;
    } else {
      this.predicate = 'id';
      this.ascending = true;
    }
    this.currentSearch = query;
    this.loadAll();
  }

  ngOnInit(): void {
    this.loadAll();
    this.registerChangeInWatchListEntries();
  }

  ngOnDestroy(): void {
    if (this.eventSubscriber) {
      this.eventManager.destroy(this.eventSubscriber);
    }
  }

  trackId(index: number, item: IWatchListEntry): number {
    // eslint-disable-next-line @typescript-eslint/no-unnecessary-type-assertion
    return item.id!;
  }

  registerChangeInWatchListEntries(): void {
    this.eventSubscriber = this.eventManager.subscribe('watchListEntryListModification', () => this.reset());
  }

  delete(watchListEntry: IWatchListEntry): void {
    const modalRef = this.modalService.open(WatchListEntryDeleteDialogComponent, { size: 'lg', backdrop: 'static' });
    modalRef.componentInstance.watchListEntry = watchListEntry;
  }

  sort(): string[] {
    const result = [this.predicate + ',' + (this.ascending ? 'asc' : 'desc')];
    if (this.predicate !== 'id') {
      result.push('id');
    }
    return result;
  }

  protected paginateWatchListEntries(data: IWatchListEntry[] | null, headers: HttpHeaders): void {
    const headersLink = headers.get('link');
    this.links = this.parseLinks.parse(headersLink ? headersLink : '');
    if (data) {
      for (const dataElem of data) {
        this.watchListEntries.push(dataElem);
      }
    }
  }
}
