package at.ac.uibk.gitsearch.service;

import static org.mockito.ArgumentMatchers.any;
import static org.mockito.ArgumentMatchers.anyString;
import static org.mockito.Mockito.doNothing;
import static org.mockito.Mockito.doReturn;
import static org.mockito.Mockito.spy;
import static org.mockito.Mockito.times;
import static org.mockito.Mockito.verify;

import at.ac.uibk.gitsearch.GitsearchApp;
import at.ac.uibk.gitsearch.es.model.ArtemisExerciseInfo;
import at.ac.uibk.gitsearch.properties.ApplicationProperties;
import at.ac.uibk.gitsearch.properties.ApplicationProperties.RegisteredConnector;
import at.ac.uibk.gitsearch.service.dto.MetadataUserDTO;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.nio.charset.StandardCharsets;
import java.nio.file.FileSystemException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.concurrent.TimeUnit;
import java.util.stream.Collectors;
import java.util.zip.ZipEntry;
import java.util.zip.ZipInputStream;
import java.util.zip.ZipOutputStream;
import org.apache.commons.io.FileUtils;
import org.apache.commons.lang3.StringUtils;
import org.eclipse.jgit.lib.Repository;
import org.eclipse.jgit.storage.file.FileRepositoryBuilder;
import org.gitlab4j.api.models.Group;
import org.junit.Assert;
import org.junit.jupiter.api.AfterEach;
import org.junit.jupiter.api.Assertions;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.Timeout;
import org.junit.runner.RunWith;
import org.mockito.InjectMocks;
import org.mockito.MockitoAnnotations;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.boot.test.mock.mockito.MockBean;
import org.springframework.security.test.context.support.WithMockUser;
import org.springframework.test.context.junit4.SpringJUnit4ClassRunner;
import org.springframework.transaction.annotation.Transactional;

/**
 * Integration tests for {@link UserWatchListService}.
 */
@SpringBootTest(classes = GitsearchApp.class)
@RunWith(SpringJUnit4ClassRunner.class)
@WithMockUser(value = "test", authorities = "sharing")
@Transactional
class ExerciseImportServiceIT {

    private final Logger log = LoggerFactory.getLogger(ExerciseImportServiceIT.class);

    @InjectMocks
    @Autowired
    ExerciseImportService exerciseImportService;

    @Autowired
    private ApplicationProperties applicationProperties;

    @MockBean
    GitlabService gitlabService;

    String token = "testToken";

    String TMP_DIR_PROPERTY = "java.io.tmpdir";

    Path testZipPath = Paths.get(System.getProperty(TMP_DIR_PROPERTY), "testZip.zip");

    ArtemisExerciseInfo preliminaryArtemisExerciseInfo = new ArtemisExerciseInfo();

    ArtemisExerciseInfo fullArtemisExerciseInfo = new ArtemisExerciseInfo();

    ArrayList<?> allowedGitlabGroupId = new ArrayList<>(Arrays.asList(Long.class, String.class, Group.class));

    @BeforeEach
    public void init() throws IOException {
        MockitoAnnotations.openMocks(this);

        preliminaryArtemisExerciseInfo.setMetadataVersion("1");
        preliminaryArtemisExerciseInfo.setTitle("TestTitle");
        preliminaryArtemisExerciseInfo.setLicense("testLicense");
        preliminaryArtemisExerciseInfo.setLearningResourceType("artemis");
        preliminaryArtemisExerciseInfo.setStructure("atomic");
        preliminaryArtemisExerciseInfo.setFormat(List.of("artemis"));

        MetadataUserDTO metadataUser = new MetadataUserDTO();
        metadataUser.setName("Test User");
        metadataUser.setAffiliation("Test Affiliation");
        metadataUser.setEmail("test@user");

        fullArtemisExerciseInfo.setMetadataVersion("1");
        fullArtemisExerciseInfo.setTitle("TestTitle");
        fullArtemisExerciseInfo.setLicense("testLicense");
        fullArtemisExerciseInfo.setLearningResourceType("artemis");
        fullArtemisExerciseInfo.setStructure("atomic");
        fullArtemisExerciseInfo.setFormat(List.of("artemis"));
        fullArtemisExerciseInfo.setKeyword(List.of("keyword"));
        fullArtemisExerciseInfo.setDescription("TestDescription");
        fullArtemisExerciseInfo.setLanguage(List.of("en"));
        fullArtemisExerciseInfo.setCreator(List.of(metadataUser));
        fullArtemisExerciseInfo.setPublisher(List.of(metadataUser));
        fullArtemisExerciseInfo.setIdentifier("https://test:8080/test/path/url");
        fullArtemisExerciseInfo.setProgrammingLanguage(List.of("C"));
    }

    @AfterEach
    public void cleanUp() throws IOException {
        if (new File(System.getProperty("java.io.tmpdir"), token).exists()) {
            FileUtils.deleteDirectory(new File(System.getProperty("java.io.tmpdir") + token));
        }
        try {
            Files.deleteIfExists(testZipPath);
        } catch (FileSystemException fse) {
            log.warn("Couldn't delete {}, if the problem persists, please remove this file manually", testZipPath.toAbsolutePath());
        }
    }

    @Test
    @Timeout(value = 5, unit = TimeUnit.MINUTES)
    void testImportExercise() throws Exception {
        Long gitlabGroup = 1L;

        File metadata = Paths.get(System.getProperty(TMP_DIR_PROPERTY), token, "Exercise-Details-something.json").toFile();
        File problemStatement = Paths.get(System.getProperty(TMP_DIR_PROPERTY), token, "Problem-Statement-something.md").toFile();

        metadata.getParentFile().mkdirs();
        metadata.createNewFile();
        problemStatement.createNewFile();
        Group group = new Group();
        Repository jgitRepo = FileRepositoryBuilder.create(Paths.get(System.getProperty("java.io.tmpdir"), token, ".git").toFile());

        // mocks
        doNothing().when(gitlabService).setJGitDefaultAuth();
        doReturn(group).when(gitlabService).getGroupById(gitlabGroup);
        doReturn("repoUrl").when(gitlabService).createRepository(group, fullArtemisExerciseInfo.getTitle());
        doReturn(jgitRepo).when(gitlabService).checkoutRepo("repoUrl");
        doNothing().when(gitlabService).commitAndPushToRepo(jgitRepo, new File(System.getProperty("java.io.tmpdir"), token));

        exerciseImportService.importExercise(fullArtemisExerciseInfo, token, gitlabGroup);

        // asserts
        verify(gitlabService, times(1)).createRepository(group, preliminaryArtemisExerciseInfo.getTitle());
        verify(gitlabService, times(1)).commitAndPushToRepo(jgitRepo, new File(System.getProperty("java.io.tmpdir"), token));
        Assert.assertTrue(allowedGitlabGroupId.contains(gitlabGroup.getClass()));
    }

    @Test
    void testApplyExerciseInfoChanges() throws Exception {
        File metadata = Paths.get(System.getProperty("java.io.tmpdir"), token, "metadata.yaml").toFile();
        metadata.getParentFile().mkdirs();
        metadata.createNewFile();

        exerciseImportService.applyExerciseInfoChanges(fullArtemisExerciseInfo, token);

        Assert.assertTrue(new String(Files.readAllBytes(metadata.toPath())).contains("learningResourceType: \"artemis\""));
        Assert.assertTrue(new String(Files.readAllBytes(metadata.toPath())).contains("license: \"testLicense\""));
        Assert.assertTrue(new String(Files.readAllBytes(metadata.toPath())).contains("metadataVersion: \"1\""));
        Assert.assertTrue(new String(Files.readAllBytes(metadata.toPath())).contains("title: \"TestTitle\""));
        Assert.assertTrue(new String(Files.readAllBytes(metadata.toPath())).contains("description: \"TestDescription\""));
        Assert.assertTrue(new String(Files.readAllBytes(metadata.toPath())).contains("language:\n - \"en\""));
        Assert.assertTrue(new String(Files.readAllBytes(metadata.toPath())).contains("creator:\n -\n  name: \"Test User\""));
    }

    @Test
    void testGetArtemisExerciseInfo() throws Exception {
        File metadata = Paths.get(System.getProperty("java.io.tmpdir"), token, "Exercise-Details-something.json").toFile();
        metadata.getParentFile().mkdirs();
        metadata.createNewFile();
        FileWriter writer = new FileWriter(metadata, StandardCharsets.UTF_8);
        writer.write(
            "{\n" +
            "  \"metadataVersion\": \"1\",\n" +
            "  \"type\": \"programming exercise\",\n" +
            "  \"title\": \"TestTitle\",\n" +
            "  \"license\": \"testLicense\",\n" +
            "  \"maxPoints\": 100,\n" +
            "  \"bonusPoints\": 100,\n" +
            "  \"mode\": \"INDIVIDUAL\",\n" +
            "  \"staticCodeAnalysis\": false,\n" +
            "  \"allowOfflineIDE\": false,\n" +
            "  \"allowOnlineEditor\": false,\n" +
            "  \"showTestNamesToStudents\": false,\n" +
            "  \"sequentialTestRuns\": false\n" +
            "}"
        );
        writer.close();

        preliminaryArtemisExerciseInfo.setMaxPoints(100f);
        preliminaryArtemisExerciseInfo.setBonusPoints(100f);
        preliminaryArtemisExerciseInfo.setMode(ArtemisExerciseInfo.ExerciseMode.INDIVIDUAL);
        preliminaryArtemisExerciseInfo.setStaticCodeAnalysis(false);
        preliminaryArtemisExerciseInfo.setAllowOfflineIDE(false);
        preliminaryArtemisExerciseInfo.setAllowOnlineEditor(false);
        preliminaryArtemisExerciseInfo.setShowTestNamesToStudents(false);
        preliminaryArtemisExerciseInfo.setSequentialTestRuns(false);
        preliminaryArtemisExerciseInfo.setStructure("atomic");

        Assert.assertEquals(preliminaryArtemisExerciseInfo, exerciseImportService.getArtemisExerciseInfo(token));
    }

    @Test
    @SuppressWarnings("PMD")
    void testGetArtemisExerciseInfo_onlyMetadata() throws Exception {
        File metadata = Paths.get(System.getProperty("java.io.tmpdir"), token, "Exercise-Details-something.json").toFile();
        metadata.getParentFile().mkdirs();
        metadata.createNewFile();
        FileWriter writer = new FileWriter(metadata, StandardCharsets.UTF_8);
        writer.write(
            "{\n" +
            "  \"metadataVersion\": \"1\",\n" +
            "  \"type\": \"programming exercise\",\n" +
            "  \"title\": \"TestTitle\",\n" +
            "  \"license\": \"testLicense\"\n" +
            "}"
        );
        writer.close();

        Assert.assertEquals(preliminaryArtemisExerciseInfo, exerciseImportService.getArtemisExerciseInfo(token));
    }

    @Test
    void testValidate() throws Exception {
        String testApiKeyCorrect = System.getenv("SHARING_PLUGIN_ACCESS_TOKEN");
        String testApiKeyIncorrect = "wrong-apikey";

        // this is a hilarious workaround for gitlab pipeline tests:
        if (StringUtils.isEmpty(testApiKeyCorrect)) {
            testApiKeyCorrect = "${CONNECTOR_ARTEMIS_TOKEN}";
        }

        String message = String.format(
            "Error: testApiKeyCorrect: %s: other keys: %s",
            testApiKeyIncorrect,
            applicationProperties
                .getRegisteredConnectors()
                .stream()
                .map(RegisteredConnector::getAccessToken)
                .collect(Collectors.joining(", "))
        );

        Assert.assertTrue(message, exerciseImportService.validate(testApiKeyCorrect));
        Assert.assertFalse(exerciseImportService.validate(testApiKeyIncorrect));
    }

    @Test
    void testValidateExerciseInfo_correct() throws Exception {
        // Should not throw any Exception
        exerciseImportService.validateExerciseInfo(fullArtemisExerciseInfo);
    }

    @Test
    void testValidateExerciseInfo_missing() throws Exception {
        Assertions.assertThrows(ArtemisImportError.class, () -> exerciseImportService.validateExerciseInfo(preliminaryArtemisExerciseInfo));
    }

    @Test
    @Timeout(value = 5, unit = TimeUnit.MINUTES)
    void testSanitizeExerciseInfo() throws Exception {
        Long gitlabGroup = 1L;

        File metadata = Paths.get(System.getProperty("java.io.tmpdir"), token, "Exercise-Details-something.json").toFile();
        metadata.getParentFile().mkdirs();
        metadata.createNewFile();
        Group group = new Group();
        Repository jgitRepo = FileRepositoryBuilder.create(Paths.get(System.getProperty("java.io.tmpdir"), token, ".git").toFile());

        // mocks
        MetadataUserDTO user = new MetadataUserDTO();
        user.setName("<script>Name");
        user.setEmail("user@<script>test");
        user.setAffiliation("test<script>affiliation");

        fullArtemisExerciseInfo.setMetadataVersion("<script>");
        fullArtemisExerciseInfo.setTitle(fullArtemisExerciseInfo.getTitle() + "<script>");
        fullArtemisExerciseInfo.setDescription(fullArtemisExerciseInfo.getDescription() + "<script>");
        fullArtemisExerciseInfo.setKeyword(List.of("Key<script>word"));
        fullArtemisExerciseInfo.setCreator(List.of(user));
        fullArtemisExerciseInfo.setPublisher(List.of(user));
        fullArtemisExerciseInfo.setLicense(fullArtemisExerciseInfo.getLicense() + "<script>");
        fullArtemisExerciseInfo.setFormat(List.of("<script>"));
        fullArtemisExerciseInfo.setLearningResourceType(fullArtemisExerciseInfo.getLearningResourceType() + "<script>");

        doNothing().when(gitlabService).setJGitDefaultAuth();
        doReturn(group).when(gitlabService).getGroupById(gitlabGroup);
        doReturn("repoUrl").when(gitlabService).createRepository(any(), anyString());
        doReturn(jgitRepo).when(gitlabService).checkoutRepo("repoUrl");
        doNothing().when(gitlabService).commitAndPushToRepo(jgitRepo, new File(System.getProperty("java.io.tmpdir"), token));

        exerciseImportService.importExercise(fullArtemisExerciseInfo, token, gitlabGroup);

        Assert.assertFalse(fullArtemisExerciseInfo.getMetadataVersion().contains("<script"));
        Assert.assertFalse(fullArtemisExerciseInfo.getTitle().contains("<script"));
        Assert.assertFalse(fullArtemisExerciseInfo.getDescription().contains("<script"));
        Assert.assertFalse(fullArtemisExerciseInfo.getKeyword().get(0).contains("<script"));
        Assert.assertFalse(fullArtemisExerciseInfo.getCreator().get(0).getName().contains("<script"));
        Assert.assertFalse(fullArtemisExerciseInfo.getCreator().get(0).getEmail().contains("<script"));
        Assert.assertFalse(fullArtemisExerciseInfo.getCreator().get(0).getAffiliation().contains("<script"));
        Assert.assertFalse(fullArtemisExerciseInfo.getPublisher().get(0).getName().contains("<script"));
        Assert.assertFalse(fullArtemisExerciseInfo.getPublisher().get(0).getEmail().contains("<script"));
        Assert.assertFalse(fullArtemisExerciseInfo.getPublisher().get(0).getAffiliation().contains("<script"));
        Assert.assertFalse(fullArtemisExerciseInfo.getLicense().contains("<script"));
        Assert.assertFalse(fullArtemisExerciseInfo.getFormat().contains("<script"));
        Assert.assertFalse(fullArtemisExerciseInfo.getLearningResourceType().contains("<script"));
    }

    @Test
    @Timeout(value = 5, unit = TimeUnit.MINUTES)
    void testImportExerciseFromArtemis() throws Exception {
        ExerciseImportService mockExerciseImportService = spy(exerciseImportService);

        ZipOutputStream out = new ZipOutputStream(new FileOutputStream(testZipPath.toFile()));
        out.putNextEntry(new ZipEntry("testFile.yml"));
        byte[] buffer = new byte[1024];

        int length;
        FileInputStream fis = new FileInputStream(testZipPath.toFile());
        while ((length = fis.read(buffer)) >= 0) {
            out.write(buffer, 0, length);
        }
        out.closeEntry();
        fis.close();

        doReturn(new ZipInputStream(new FileInputStream(testZipPath.toFile())))
            .when(mockExerciseImportService)
            .getZipInputStreamFromUrl(anyString());
        mockExerciseImportService.importExerciseFromArtemis(token);

        Assert.assertTrue(Paths.get(System.getProperty(TMP_DIR_PROPERTY), token).toFile().exists());
    }
}
