package at.ac.uibk.gitsearch.service;

import static org.assertj.core.api.Assertions.assertThat;
import static org.hamcrest.Matchers.anyOf;
import static org.hamcrest.Matchers.contains;
import static org.hamcrest.Matchers.containsString;
import static org.hamcrest.Matchers.containsStringIgnoringCase;
import static org.hamcrest.Matchers.everyItem;
import static org.hamcrest.Matchers.hasItemInArray;
import static org.hamcrest.Matchers.hasProperty;
import static org.hamcrest.Matchers.is;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.ArgumentMatchers.anyString;
import static org.mockito.ArgumentMatchers.eq;
import static org.mockito.Mockito.doReturn;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.when;

import at.ac.uibk.gitsearch.IntegrationTest;
import at.ac.uibk.gitsearch.repository.gitlab.GitLabRepository;
import at.ac.uibk.gitsearch.repository.jpa.StatisticsRepository;
import at.ac.uibk.gitsearch.repository.search.ElasticSearchRepository;
import at.ac.uibk.gitsearch.repository.search.MetaDataRepository;
import at.ac.uibk.gitsearch.repository.search.testESService.ElasticSearchTestConfiguration;
import at.ac.uibk.gitsearch.security.jwt.TokenProvider;
import at.ac.uibk.gitsearch.service.PluginManagementService.ConnectorConfigWrapper;
import at.ac.uibk.gitsearch.service.dto.AutoCompleteEntry;
import at.ac.uibk.gitsearch.service.dto.StatisticsDTO;
import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Comparator;
import java.util.List;
import java.util.Optional;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.stream.Collectors;
import javax.ws.rs.NotFoundException;
import org.codeability.sharing.plugins.api.SharingPluginConfig;
import org.codeability.sharing.plugins.api.search.SearchInputDTO;
import org.codeability.sharing.plugins.api.search.SearchInputMetadataDTO;
import org.codeability.sharing.plugins.api.search.SearchResultDTO;
import org.codeability.sharing.plugins.api.search.SearchResultsDTO;
import org.elasticsearch.node.NodeValidationException;
import org.gitlab4j.api.GitLabApi;
import org.gitlab4j.api.GitLabApiException;
import org.gitlab4j.api.RepositoryApi;
import org.hamcrest.MatcherAssert;
import org.junit.Assert;
import org.junit.jupiter.api.AfterAll;
import org.junit.jupiter.api.Assertions;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.Timeout;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.mock.mockito.MockBean;
import org.springframework.security.test.context.support.WithAnonymousUser;
import org.springframework.security.test.context.support.WithMockUser;
import org.springframework.transaction.annotation.Transactional;

@IntegrationTest
@WithMockUser(value = "test", authorities = "sharing")
public class SearchServiceIT {

    @MockBean
    PluginManagementService pluginManagementService;

    @MockBean
    TokenProvider tokenProvider;

    @MockBean
    GitLabRepository gitLabRepository;

    @Autowired
    MetaDataRepository metaDataRepository;

    @Autowired
    StatisticsRepository statisticsRepository;

    @Autowired
    StatisticsService statisticsService;

    @Autowired
    ElasticSearchRepository elasticSearchRepository;

    @MockBean
    RepositoryApi mockedRepositoryApi;

    private static final Logger LOGGER = LoggerFactory.getLogger(SearchServiceIT.class);

    @Autowired
    private SearchService searchService;

    @Autowired
    ElasticSearchTestConfiguration elasticSearchTestConfiguration;

    static ElasticSearchTestConfiguration staticElasticSearchTestConfiguration;

    @BeforeEach // must be started as BeforeEach, in order to autowire the
    // ElasticSearchTestConfiguration
    @Timeout(value = 3, unit = TimeUnit.MINUTES)
    public void setUpESServer() throws IOException, NodeValidationException {
        if (staticElasticSearchTestConfiguration == null) {
            elasticSearchTestConfiguration.startTestNode();
            staticElasticSearchTestConfiguration = elasticSearchTestConfiguration;
        }
    }

    @AfterAll
    @Timeout(value = 2, unit = TimeUnit.MINUTES)
    public static void shutDownESServer() {
        if (staticElasticSearchTestConfiguration != null) {
            staticElasticSearchTestConfiguration.stopTestContainer();
            staticElasticSearchTestConfiguration = null;
        }
    }

    @BeforeEach
    public void initMock() {
        Collection<ConnectorConfigWrapper> mockConfigs = new ArrayList<>();
        mockConfigs.add(
            new ConnectorConfigWrapper(
                new SharingPluginConfig(
                    "mockedPlugin",
                    new SharingPluginConfig.Action[] {
                        new SharingPluginConfig.Action("mockedAction", "http://unused/unused-transferURL", "mocked Action", "true"),
                    }
                ),
                "http://unused/hopefully-unused"
            )
        );

        mockConfigs.add(
            new ConnectorConfigWrapper(
                new SharingPluginConfig(
                    "Artemis Sharing Connector",
                    new SharingPluginConfig.Action[] {
                        new SharingPluginConfig.Action("Import", "http://unused/unused-transferURL", "Export to Artemis", "true"),
                    }
                ),
                "http://unused/hopefully-unused"
            )
        );

        when(pluginManagementService.getRegisteredPluginConfigs()).thenReturn(mockConfigs);
    }

    @Test
    @Timeout(value = 1, unit = TimeUnit.MINUTES)
    void testAllSearch() throws Exception {
        final SearchInputMetadataDTO searchMetadata = new SearchInputMetadataDTO(null, null, null, null, null, null);
        SearchInputDTO searchQuery = new SearchInputDTO(null, searchMetadata, null, null, null, 0);
        SearchResultsDTO searchResultPage = searchService.searchResultPage(searchQuery);

        Assert.assertTrue("At least one hit?", searchResultPage.getSearchResult().size() >= 1);
        Assert.assertEquals("We start at 0", 0, searchResultPage.getPageStartIndex());
        LOGGER.info("found {} hits for all", searchResultPage.getHitCount());
    }

    @Test
    @Timeout(value = 10, unit = TimeUnit.MINUTES)
    @Transactional
    void testAllSearchWithStatistics() throws Exception {
        // prepare
        final SearchInputMetadataDTO searchMetadata = new SearchInputMetadataDTO(null, null, null, null, null, null);
        SearchInputDTO searchQuery = new SearchInputDTO(null, searchMetadata, null, null, null, 0);
        searchQuery.setPageSize(500);
        SearchResultsDTO searchResultPage = searchService.searchResultPage(searchQuery);

        AtomicInteger downloadsCount = new AtomicInteger(0);
        AtomicBoolean badgeRewarded = new AtomicBoolean(false);
        AtomicInteger viewsCount = new AtomicInteger(0);
        // setup statistics data
        searchResultPage
            .getSearchResult()
            .stream()
            .map(sr -> {
                assertThat(sr.getSearchStatistics()).isNull(); // this may be critical, if some other test already
                return sr;
            })
            .skip(1) // leave one stat empty
            .forEach(sr -> {
                // define statistics
                StatisticsDTO stat = new StatisticsDTO();
                stat.setExerciseID(sr.getExerciseId());
                stat.setDownloads(5 + downloadsCount.get());
                stat.setBadgeRewarded(badgeRewarded.get());
                stat.setViews(95 + (viewsCount.get()));
                if (badgeRewarded.getAndSet(!badgeRewarded.get())) {
                    downloadsCount.getAndIncrement();
                }
                viewsCount.getAndIncrement();
                statisticsService.save(stat);
            });

        statisticsRepository.flush();

        elasticSearchRepository.waitForIndexingFinished();
        SearchResultsDTO searchResultPage2 = searchService.searchResultPage(searchQuery);
        // it occurs that upon requesting elastic search right after updating the
        // indices, the old order is retrieved

        Comparator<SearchResultDTO> orderComparator = Comparator.comparing(sr -> {
            if (sr.getSearchStatistics() == null) {
                return 0D;
            }
            return (
                (sr.getSearchStatistics().getBadgeRewarded() ? 1 : 0) *
                searchQuery.getOrdering().getFactorBadge() +
                sr.getSearchStatistics().getDownloads() *
                searchQuery.getOrdering().getFactorDownloads() +
                sr.getSearchStatistics().getViews() *
                searchQuery.getOrdering().getFactorViews()
            );
        });

        SearchResultDTO previousSearchResult = null;
        var searchResultsIterator = searchResultPage2.getSearchResult().iterator();
        while (searchResultsIterator.hasNext()) {
            var sr = searchResultsIterator.next();
            if (!searchResultsIterator.hasNext()) {
                assertThat(sr.getSearchStatistics()).isNull();
            } else {
                assertThat(sr.getSearchStatistics()).isNotNull();
                assertThat(sr.getSearchStatistics().getDownloads()).isGreaterThanOrEqualTo(5);
                assertThat(sr.getSearchStatistics().getViews()).isGreaterThanOrEqualTo(95);
            }
            if (previousSearchResult != null) {
                assertThat(orderComparator.compare(previousSearchResult, sr)).isPositive();
            }
            previousSearchResult = sr;
        }
    }

    @Test
    @Timeout(value = 1, unit = TimeUnit.MINUTES)
    void testFullTextSearch() throws Exception {
        LOGGER.info("starting testFullTextSearch");
        final String wien = "Wien";
        final SearchInputMetadataDTO searchMetadata = new SearchInputMetadataDTO(null, null, null, null, null, null);
        SearchInputDTO searchQuery = new SearchInputDTO(wien, searchMetadata, null, null, null, 0);
        SearchResultsDTO searchResultPage = searchService.searchResultPage(searchQuery);

        org.junit.Assert.assertNotNull(searchResultPage.getSearchResult());
        org.junit.Assert.assertTrue("At least one test hit", searchResultPage.getHitCount() >= 1);
        MatcherAssert.assertThat(
            searchResultPage.getSearchResult(),
            everyItem(
                anyOf(
                    hasProperty("metadata", hasProperty("description", containsString(wien))),
                    hasProperty("metadata", hasProperty("title", containsString(wien)))
                    // hasProperty("keywords", containsString("Wien"))
                )
            )
        );

        LOGGER.info("exiting testFullTextSearch");
    }

    @Test
    @Timeout(value = 1, unit = TimeUnit.MINUTES)
    void testSearchByAutor() throws Exception {
        final String podlipnig = "Stefan Podlipnig";
        final SearchInputMetadataDTO searchMetadata = new SearchInputMetadataDTO(null, null, null, null, podlipnig, null);
        SearchInputDTO searchQuery = new SearchInputDTO(null, searchMetadata, null, null, null, 0);
        SearchResultsDTO searchResultPage = searchService.searchResultPage(searchQuery);

        org.junit.Assert.assertNotNull(searchResultPage.getSearchResult());
        org.junit.Assert.assertTrue("At least one test hit", searchResultPage.getHitCount() >= 1);
        MatcherAssert.assertThat(
            searchResultPage.getSearchResult(),
            everyItem(hasProperty("metadata", hasProperty("creator", hasItemInArray(hasProperty("name", containsString(podlipnig))))))
        );
    }

    @Test
    @Timeout(value = 1, unit = TimeUnit.MINUTES)
    void testCreatorSearch() throws IOException {
        final SearchInputMetadataDTO searchMetadata = new SearchInputMetadataDTO(null, null, null, null, "Podlipnig", null);
        SearchInputDTO searchQuery = new SearchInputDTO(null, searchMetadata, null, null, null, 0);
        SearchResultsDTO searchResultPage = searchService.searchResultPage(searchQuery);

        org.junit.Assert.assertNotNull(searchResultPage.getSearchResult());
        org.junit.Assert.assertTrue("At least one test hit", searchResultPage.getHitCount() >= 1);
        MatcherAssert.assertThat(
            searchResultPage.getSearchResult(),
            everyItem(hasProperty("metadata", hasProperty("creator", hasItemInArray(hasProperty("name", containsString("Podlipnig"))))))
        );
    }

    @Test
    @Timeout(value = 1, unit = TimeUnit.MINUTES)
    void testKeywordSearch() throws IOException {
        final SearchInputMetadataDTO searchMetadata = new SearchInputMetadataDTO(null, "latex", null, null, null, null);
        org.codeability.sharing.plugins.api.search.SearchInputDTO searchQuery = new SearchInputDTO(
            null,
            searchMetadata,
            null,
            null,
            null,
            0
        );
        SearchResultsDTO searchResultPage = searchService.searchResultPage(searchQuery);

        org.junit.Assert.assertNotNull(searchResultPage.getSearchResult());
        org.junit.Assert.assertTrue("At least one test hit", searchResultPage.getHitCount() >= 1);
        MatcherAssert.assertThat(
            searchResultPage.getSearchResult(),
            everyItem(hasProperty("metadata", hasProperty("keyword", hasItemInArray(containsString("latex")))))
        );
    }

    @Test
    @Timeout(value = 1, unit = TimeUnit.MINUTES)
    void testFormatSearch() throws IOException {
        final SearchInputMetadataDTO searchMetadata = new SearchInputMetadataDTO(null, null, null, null, null, "latex");
        SearchInputDTO searchQuery = new SearchInputDTO(null, searchMetadata, null, null, null, 0);
        SearchResultsDTO searchResultPage = searchService.searchResultPage(searchQuery);

        org.junit.Assert.assertNotNull(searchResultPage.getSearchResult());
        org.junit.Assert.assertTrue("At least one test hit", searchResultPage.getHitCount() >= 1);
        MatcherAssert.assertThat(
            searchResultPage.getSearchResult(),
            everyItem(hasProperty("metadata", hasProperty("format", hasItemInArray(containsString("latex")))))
        );
    }

    // does not work with metadata version 0.4
    // @Test
    // void testTypesSearch() throws IOException {
    // final SearchInputMetadataDTO searchMetadata = new
    // SearchInputMetadataDTO(null, "latex", null, null, null,null);
    // searchMetadata.setTypes(Collections.singletonList(ExerciseType.COLLECTION));
    // SearchInputDTO searchQuery = new SearchInputDTO(null, searchMetadata, null,
    // null, null, 0);
    // SearchResultsDTO searchResultPage =
    // searchService.searchResultPage(searchQuery);
    //
    // org.junit.Assert.assertNotNull(searchResultPage.getSearchResult());
    // org.junit.Assert.assertTrue("At least one test hit",
    // searchResultPage.getHitCount() >= 1);
    // assertThat(searchResultPage.getSearchResult(),
    // everyItem(hasProperty("metadata", hasProperty("keyword",
    // hasItemInArray(containsString("latex"))))));
    //
    // }

    @Test
    @Timeout(value = 1, unit = TimeUnit.MINUTES)
    void testNaturalLanguagesSearch() throws IOException {
        final SearchInputMetadataDTO searchMetadata = new SearchInputMetadataDTO(null, "latex", null, null, null, null);
        searchMetadata.setNaturalLanguage(List.of("de"));
        SearchInputDTO searchQuery = new SearchInputDTO(null, searchMetadata, null, null, null, 0);
        SearchResultsDTO searchResultPage = searchService.searchResultPage(searchQuery);
        org.junit.Assert.assertNotNull(searchResultPage.getSearchResult());

        org.junit.Assert.assertTrue("At least one test hit", searchResultPage.getHitCount() >= 1);
        MatcherAssert.assertThat(
            searchResultPage.getSearchResult(),
            everyItem(hasProperty("metadata", hasProperty("language", hasItemInArray(containsString("de")))))
        );
    }

    @Test
    @Timeout(value = 1, unit = TimeUnit.MINUTES)
    void testProgrammingLanguagesSearch() throws IOException {
        final SearchInputMetadataDTO searchMetadata = new SearchInputMetadataDTO("Java", null, null, null, null, null);
        SearchInputDTO searchQuery = new SearchInputDTO(null, searchMetadata, null, null, null, 0);
        SearchResultsDTO searchResultPage = searchService.searchResultPage(searchQuery);
        org.junit.Assert.assertNotNull(searchResultPage.getSearchResult());

        org.junit.Assert.assertTrue("At least one test hit", searchResultPage.getHitCount() >= 1);
        MatcherAssert.assertThat(
            searchResultPage.getSearchResult(),
            everyItem(hasProperty("metadata", hasProperty("programmingLanguage", hasItemInArray(containsStringIgnoringCase("Java")))))
        );
    }

    @Test
    @Timeout(value = 1, unit = TimeUnit.MINUTES)
    void testProgrammingLanguagesPrefixSearch() throws IOException {
        final SearchInputMetadataDTO searchMetadata = new SearchInputMetadataDTO("Jav", null, null, null, null, null);
        SearchInputDTO searchQuery = new SearchInputDTO(null, searchMetadata, null, null, null, 0);
        SearchResultsDTO searchResultPage = searchService.searchResultPage(searchQuery);
        org.junit.Assert.assertNotNull(searchResultPage.getSearchResult());

        org.junit.Assert.assertTrue("At least one test hit", searchResultPage.getHitCount() >= 1);
        MatcherAssert.assertThat(
            searchResultPage.getSearchResult(),
            everyItem(hasProperty("metadata", hasProperty("programmingLanguage", hasItemInArray(containsStringIgnoringCase("Java")))))
        );
    }

    // does not work with metadata version 0.4
    // @Test
    // void testTypesSearchNegative() throws IOException {
    // final SearchInputMetadataDTO searchMetadata = new
    // SearchInputMetadataDTO(null, "latex", null, null, null,null);
    // searchMetadata.setTypes(Collections.singletonList(ExerciseType.OTHER));
    // SearchInputDTO searchQuery = new SearchInputDTO(null, searchMetadata, null,
    // null, null, 0);
    // SearchResultsDTO searchResultPage =
    // searchService.searchResultPage(searchQuery);
    //
    // org.junit.Assert.assertNotNull(searchResultPage.getSearchResult());
    // org.junit.Assert.assertTrue("We expect no hit for \"latex\" and
    // ExerciseType.OTHER", searchResultPage.getHitCount() == 0);
    //
    // }

    // @Test()
    // @Ignore() // Test funktioniert momentan nicht?
    void testLicenseSearch() throws IOException {
        final SearchInputMetadataDTO searchMetadata = new SearchInputMetadataDTO(null, null, null, "MIT", null, null);
        SearchInputDTO searchQuery = new SearchInputDTO(null, searchMetadata, null, null, null, 0);
        SearchResultsDTO searchResultPage = searchService.searchResultPage(searchQuery);

        org.junit.Assert.assertNotNull(searchResultPage.getSearchResult());
        org.junit.Assert.assertTrue("At least one test hit", searchResultPage.getHitCount() >= 1);
        MatcherAssert.assertThat(
            searchResultPage.getSearchResult(),
            everyItem(hasProperty("metadata", hasProperty("license", containsString("MIT"))))
        );
    }

    @Test
    @Timeout(value = 1, unit = TimeUnit.MINUTES)
    void testKeywordAutocompletion() throws IOException {
        final List<AutoCompleteEntry> keywordsAutoComplete = metaDataRepository.getKeywordsAutoComplete("Jav", 10);
        MatcherAssert.assertThat(keywordsAutoComplete, contains(hasProperty("target", is("Java"))));
    }

    @Test
    @Timeout(value = 1, unit = TimeUnit.MINUTES)
    void testCreatorAutocompletion() throws IOException {
        final List<AutoCompleteEntry> creatorAutoComplete = metaDataRepository.getCreatorAutoComplete("Pod", 10);
        MatcherAssert.assertThat(creatorAutoComplete, contains(hasProperty("target", is("Stefan Podlipnig"))));
        final List<AutoCompleteEntry> creatorAutoComplete2 = metaDataRepository.getCreatorAutoComplete("Po", 10);
        MatcherAssert.assertThat(creatorAutoComplete2, contains(hasProperty("target", is("Stefan Podlipnig"))));
    }

    @Test
    @Timeout(value = 1, unit = TimeUnit.MINUTES)
    void testContributorAutocompletion() throws IOException {
        final List<AutoCompleteEntry> contributorAutoComplete = metaDataRepository.getContributorAutoComplete("Bast", 10);
        MatcherAssert.assertThat(contributorAutoComplete, contains(hasProperty("target", is("Daniel Bastta"))));
    }

    @Test
    @Timeout(value = 1, unit = TimeUnit.MINUTES)
    void testContributorCreatorAutocompletion() throws IOException {
        final List<AutoCompleteEntry> contributorAutoComplete = metaDataRepository.getContributorCreatorAutoComplete("Bast", 10);
        MatcherAssert.assertThat(contributorAutoComplete, contains(hasProperty("target", is("Daniel Bastta"))));
    }

    @SuppressWarnings("unchecked")
    @Test
    @Timeout(value = 1, unit = TimeUnit.MINUTES)
    void testArtemisExportActionFilteringForAuthorizedUser() throws IOException, GitLabApiException {
        GitLabApi mockedGitLabApi = mock(GitLabApi.class);
        when(gitLabRepository.getGitLabApi(any(Optional.class))).thenReturn(mockedGitLabApi);
        doReturn(mockedRepositoryApi).when(mockedGitLabApi).getRepositoryApi();
        doReturn(new ByteArrayInputStream(" ".getBytes()))
            .when(mockedRepositoryApi)
            .getRepositoryArchive(eq(333), anyString(), anyString());
        doReturn(Optional.of(new TokenProvider.GitLabAccessInfo("xx", Optional.empty()))).when(tokenProvider).getGitLabAccessInfo();

        final SearchInputMetadataDTO searchMetadata = new SearchInputMetadataDTO(null, null, null, null, null, null);
        SearchInputDTO searchQuery = new SearchInputDTO(null, searchMetadata, null, null, null, 0);
        searchQuery.setPageSize(100);
        SearchResultsDTO results = searchService.searchResultPage(searchQuery);

        Optional<SearchResultDTO> result = results
            .getSearchResult()
            .stream()
            .filter(entry -> entry.getExerciseId().equals("[333]"))
            .findFirst();

        if (result.isEmpty()) {
            throw new NotFoundException("Exercise with publicVisibility and id '333' couldn't be found in the test data");
        }

        Assertions.assertTrue(
            result.get().getSupportedActions().stream().anyMatch(action -> action.getPlugin().equals("Artemis Sharing Connector"))
        );
    }

    @Test
    @Timeout(value = 1, unit = TimeUnit.MINUTES)
    @WithMockUser(value = "User without authorities")
    void testArtemisExportActionFilteringForUnauthorizedUser() throws IOException {
        final SearchInputMetadataDTO searchMetadata = new SearchInputMetadataDTO(null, null, null, null, null, null);
        SearchInputDTO searchQuery = new SearchInputDTO(null, searchMetadata, null, null, null, 0);
        searchQuery.setPageSize(100);
        SearchResultsDTO results = searchService.searchResultPage(searchQuery);

        Optional<SearchResultDTO> result = results
            .getSearchResult()
            .stream()
            .filter(entry -> entry.getExerciseId().equals("[333]"))
            .findFirst();

        if (result.isEmpty()) {
            throw new NotFoundException("Exercise with publicVisibility and id '333' couldn't be found in the test data");
        }

        Assertions.assertFalse(
            result.get().getSupportedActions().stream().anyMatch(action -> action.getPlugin().equals("Artemis Sharing Connector"))
        );
    }

    @Test
    @Timeout(value = 1, unit = TimeUnit.MINUTES)
    @WithAnonymousUser
    void testSearchStreamAllPublic() throws IOException {
        final List<SearchResultDTO> allPublicSearchResults = searchService.getAllPublicResultsAsStream().collect(Collectors.toList());

        Assert.assertTrue("At least one hit?", allPublicSearchResults.size() >= 1);
        Assert.assertTrue("At least more than one page full?", allPublicSearchResults.size() > 10);
    }

    @Test
    @Timeout(value = 1, unit = TimeUnit.MINUTES)
    @WithAnonymousUser
    void testSearchStreamAll() throws IOException {
        final List<SearchResultDTO> allSearchResults = searchService.getAllResultsAsStream().collect(Collectors.toList());
        final List<SearchResultDTO> allPublicSearchResults = searchService.getAllPublicResultsAsStream().collect(Collectors.toList());

        Assert.assertTrue("At least one hit?", allSearchResults.size() >= 1);
        Assert.assertTrue("At least more than one page full?", allSearchResults.size() > 10);

        Assert.assertTrue("At at least one non public result?", allSearchResults.size() > allPublicSearchResults.size());
    }
}
