package at.ac.uibk.gitsearch.web.rest;

import static org.assertj.core.api.Assertions.assertThat;
import static org.hamcrest.Matchers.hasItem;
import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.springframework.security.test.web.servlet.request.SecurityMockMvcRequestPostProcessors.csrf;
import static org.springframework.test.web.servlet.request.MockMvcRequestBuilders.delete;
import static org.springframework.test.web.servlet.request.MockMvcRequestBuilders.get;
import static org.springframework.test.web.servlet.request.MockMvcRequestBuilders.patch;
import static org.springframework.test.web.servlet.request.MockMvcRequestBuilders.post;
import static org.springframework.test.web.servlet.request.MockMvcRequestBuilders.put;
import static org.springframework.test.web.servlet.result.MockMvcResultHandlers.print;
import static org.springframework.test.web.servlet.result.MockMvcResultMatchers.content;
import static org.springframework.test.web.servlet.result.MockMvcResultMatchers.jsonPath;
import static org.springframework.test.web.servlet.result.MockMvcResultMatchers.status;

import at.ac.uibk.gitsearch.IntegrationTest;
import at.ac.uibk.gitsearch.domain.Review;
import at.ac.uibk.gitsearch.domain.ReviewComment;
import at.ac.uibk.gitsearch.domain.ReviewRating;
import at.ac.uibk.gitsearch.domain.User;
import at.ac.uibk.gitsearch.domain.enumeration.ReviewStatus;
import at.ac.uibk.gitsearch.repository.jpa.ReviewRatingRepository;
import at.ac.uibk.gitsearch.security.AuthoritiesConstants;
import at.ac.uibk.gitsearch.service.dto.ReviewRatingDTO;
import at.ac.uibk.gitsearch.service.mapper.ReviewRatingMapper;
import java.util.List;
import java.util.Random;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicLong;
import javax.persistence.EntityManager;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.Timeout;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.autoconfigure.web.servlet.AutoConfigureMockMvc;
import org.springframework.http.MediaType;
import org.springframework.security.test.context.support.WithMockUser;
import org.springframework.test.web.servlet.MockMvc;
import org.springframework.transaction.annotation.Transactional;

/**
 * Integration tests for the {@link ReviewRatingResource} REST controller.
 */
@IntegrationTest
@WithMockUser(authorities = AuthoritiesConstants.ADMIN)
@AutoConfigureMockMvc
class ReviewRatingResourceIT {

    private static final ReviewStatus DEFAULT_STATUS = ReviewStatus.REVIEW_IN_PROGRESS;
    private static final ReviewStatus UPDATED_STATUS = ReviewStatus.IMPROVEMENTS_REQUESTED;

    private static final String ENTITY_API_URL = "/api/review-ratings";
    private static final String ENTITY_API_URL_ID = ENTITY_API_URL + "/{id}";

    private static Random random = new Random();
    private static AtomicLong count = new AtomicLong(random.nextInt() + (2 * Integer.MAX_VALUE));

    @Autowired
    private ReviewRatingRepository reviewRatingRepository;

    @Autowired
    private ReviewRatingMapper reviewRatingMapper;

    @Autowired
    private EntityManager em;

    @Autowired
    private MockMvc restReviewRatingMockMvc;

    private ReviewRating reviewRating;

    private ReviewRating updatedReviewRating;

    private static ReviewComment sampleComment;

    private ReviewComment updatedComment;

    /**
     * Create an entity for this test.
     *
     * This is a static method, as tests for other entities might also need it,
     * if they test an entity which requires the current entity.
     */
    public static ReviewRating createEntity(EntityManager em) {
        Review review = ReviewResourceIT.createEntity(em);
        // Add required entity
        User user = UserResourceIT.createEntity(em);
        em.persist(user);
        em.flush();

        sampleComment = new ReviewComment();

        sampleComment.setDescriptionIsClear(false);
        sampleComment.setRequirementsAreClear(false);
        sampleComment.setDescriptionComment("descriptionComment");

        sampleComment.setInformationRequiredComplete(false);
        sampleComment.setAllMaterialsAvailable(false);
        sampleComment.setInformationRequiredComment("informationRequiredComment");

        sampleComment.setIsStructured(false);
        sampleComment.setSubExercisesAreClear(false);
        sampleComment.setStructuredComment("structuredComment");

        sampleComment.setSolutionIsCorrect(false);
        sampleComment.setSolutionComment("solutionComment");

        sampleComment.setMetadataIsComplete(false);
        sampleComment.setMetadataIsCorrect(false);
        sampleComment.setMetadataComment("metadataComment");

        ReviewRating reviewRating = new ReviewRating();
        reviewRating.setStatus(DEFAULT_STATUS);
        reviewRating.setReview(review);
        reviewRating.setUser(user);
        reviewRating.setComment(sampleComment);

        return reviewRating;
    }

    /**
     * Create an updated entity for this test.
     *
     * This is a static method, as tests for other entities might also need it,
     * if they test an entity which requires the current entity.
     */
    public ReviewRating createUpdatedEntity(EntityManager em) {
        updatedComment = new ReviewComment();

        updatedComment.setDescriptionIsClear(true); // Changed this to true
        updatedComment.setRequirementsAreClear(true); // Changed this to true
        updatedComment.setDescriptionComment("The description clear and the requirements are now clear.");

        updatedComment.setInformationRequiredComplete(true); // Changed this to true
        updatedComment.setAllMaterialsAvailable(true); // Changed this to true
        updatedComment.setInformationRequiredComment("Some required information is missing, but all materials are now available.");

        updatedComment.setIsStructured(true); // Changed this to true
        updatedComment.setSubExercisesAreClear(true); // Changed this to true
        updatedComment.setStructuredComment("The structure is not good anymore, but the sub-exercises are now clear.");

        updatedComment.setSolutionIsCorrect(true); // Changed this to true
        updatedComment.setSolutionComment("The solution is not correct anymore.");

        updatedComment.setMetadataIsComplete(true); // Changed this to true
        updatedComment.setMetadataIsCorrect(true); // Changed this to true
        updatedComment.setMetadataComment("The metadata is complete but not correct anymore.");

        updatedReviewRating = new ReviewRating();
        updatedReviewRating.setComment(updatedComment);
        updatedReviewRating.setStatus(UPDATED_STATUS);
        return updatedReviewRating;
    }

    @BeforeEach
    public void initTest() {
        reviewRatingRepository.deleteAll();

        reviewRating = createEntity(em);
    }

    @Test
    @Transactional
    void createReviewRating() throws Exception {
        int databaseSizeBeforeCreate = reviewRatingRepository.findAll().size();
        // Create the ReviewRating
        ReviewRatingDTO reviewRatingDTO = reviewRatingMapper.toDto(reviewRating);
        restReviewRatingMockMvc
            .perform(
                post(ENTITY_API_URL)
                    .contentType(MediaType.APPLICATION_JSON)
                    .content(TestUtil.convertObjectToJsonBytes(reviewRatingDTO))
                    .with(csrf().asHeader())
            )
            .andDo(print())
            .andExpect(status().isCreated());

        // Validate the ReviewRating in the database
        List<ReviewRating> reviewRatingList = reviewRatingRepository.findAll();
        assertThat(reviewRatingList).hasSize(databaseSizeBeforeCreate + 1);
        ReviewRating testReviewRating = reviewRatingList.get(reviewRatingList.size() - 1);

        // to ensure equal, with have to set the database id
        sampleComment.setId(testReviewRating.getComment().getId());
        assertThat(testReviewRating.getComment()).isEqualTo(sampleComment);
        assertThat(testReviewRating.getStatus()).isEqualTo(DEFAULT_STATUS);
    }

    @Test
    @Transactional
    void createReviewRatingWithExistingId() throws Exception {
        // Create the ReviewRating with an existing ID
        reviewRating.setId(1L);
        ReviewRatingDTO reviewRatingDTO = reviewRatingMapper.toDto(reviewRating);

        int databaseSizeBeforeCreate = reviewRatingRepository.findAll().size();

        // An entity with an existing ID cannot be created, so this API call must fail
        restReviewRatingMockMvc
            .perform(
                post(ENTITY_API_URL)
                    .contentType(MediaType.APPLICATION_JSON)
                    .content(TestUtil.convertObjectToJsonBytes(reviewRatingDTO))
                    .with(csrf().asHeader())
            )
            .andExpect(status().isBadRequest());

        // Validate the ReviewRating in the database
        List<ReviewRating> reviewRatingList = reviewRatingRepository.findAll();
        assertThat(reviewRatingList).hasSize(databaseSizeBeforeCreate);
    }

    @Test
    @Transactional
    void getAllReviewRatings() throws Exception {
        // Initialize the database
        reviewRatingRepository.saveAndFlush(reviewRating);

        ReviewRating reviewRating = reviewRatingRepository.findAll().get(0);

        // Get all the reviewRatingList
        restReviewRatingMockMvc
            .perform(get(ENTITY_API_URL + "?sort=id,desc"))
            .andExpect(status().isOk())
            .andExpect(content().contentType(MediaType.APPLICATION_JSON_VALUE))
            .andExpect(jsonPath("$.[*].id").value(hasItem(reviewRating.getId().intValue())))
            .andExpect(jsonPath("$.[*].comment.descriptionIsClear").value(hasItem(reviewRating.getComment().getDescriptionIsClear())))
            .andExpect(jsonPath("$.[*].comment.requirementsAreClear").value(hasItem(reviewRating.getComment().getRequirementsAreClear())))
            .andExpect(jsonPath("$.[*].status").value(hasItem(DEFAULT_STATUS.toString())));
    }

    @Test
    @Transactional
    void getReviewRating() throws Exception {
        // Initialize the database
        reviewRatingRepository.saveAndFlush(reviewRating);

        ReviewRating reviewRating = reviewRatingRepository.findAll().get(0);

        // Get the reviewRating
        restReviewRatingMockMvc
            .perform(get(ENTITY_API_URL_ID, reviewRating.getId()))
            .andExpect(status().isOk())
            .andExpect(content().contentType(MediaType.APPLICATION_JSON_VALUE))
            .andExpect(jsonPath("$.id").value(reviewRating.getId().intValue()))
            .andExpect(jsonPath("$.comment").value(reviewRating.getComment()))
            .andExpect(jsonPath("$.status").value(DEFAULT_STATUS.toString()));
    }

    @Test
    @Transactional
    void getNonExistingReviewRating() throws Exception {
        // Get the reviewRating
        restReviewRatingMockMvc.perform(get(ENTITY_API_URL_ID, Long.MAX_VALUE)).andExpect(status().isNotFound());
    }

    @Test
    @Transactional
    void putNewReviewRating() throws Exception {
        // Initialize the database
        reviewRatingRepository.saveAndFlush(reviewRating);

        int databaseSizeBeforeUpdate = reviewRatingRepository.findAll().size();

        // Update the reviewRating
        ReviewRating updatedReviewRating = reviewRatingRepository.findById(reviewRating.getId()).get();
        // Disconnect from session so that the updates on updatedReviewRating are not directly saved in db
        em.detach(updatedReviewRating);
        updatedReviewRating.setComment(updatedComment);
        updatedReviewRating.status(UPDATED_STATUS);
        ReviewRatingDTO reviewRatingDTO = reviewRatingMapper.toDto(updatedReviewRating);

        restReviewRatingMockMvc
            .perform(
                put(ENTITY_API_URL_ID, reviewRatingDTO.getId())
                    .contentType(MediaType.APPLICATION_JSON)
                    .content(TestUtil.convertObjectToJsonBytes(reviewRatingDTO))
                    .with(csrf().asHeader())
            )
            .andExpect(status().isOk());

        // Validate the ReviewRating in the database
        List<ReviewRating> reviewRatingList = reviewRatingRepository.findAll();
        assertThat(reviewRatingList).hasSize(databaseSizeBeforeUpdate);
        ReviewRating testReviewRating = reviewRatingList.get(reviewRatingList.size() - 1);
        assertThat(testReviewRating.getComment()).isEqualTo(updatedComment);
        assertThat(testReviewRating.getStatus()).isEqualTo(UPDATED_STATUS);
    }

    @Test
    @Transactional
    void putNonExistingReviewRating() throws Exception {
        int databaseSizeBeforeUpdate = reviewRatingRepository.findAll().size();
        reviewRating.setId(count.incrementAndGet());

        // Create the ReviewRating
        ReviewRatingDTO reviewRatingDTO = reviewRatingMapper.toDto(reviewRating);

        // If the entity doesn't have an ID, it will throw BadRequestAlertException
        restReviewRatingMockMvc
            .perform(
                put(ENTITY_API_URL_ID, reviewRatingDTO.getId())
                    .contentType(MediaType.APPLICATION_JSON)
                    .content(TestUtil.convertObjectToJsonBytes(reviewRatingDTO))
                    .with(csrf().asHeader())
            )
            .andExpect(status().isBadRequest());

        // Validate the ReviewRating in the database
        List<ReviewRating> reviewRatingList = reviewRatingRepository.findAll();
        assertThat(reviewRatingList).hasSize(databaseSizeBeforeUpdate);
    }

    @Test
    @Transactional
    void putWithIdMismatchReviewRating() throws Exception {
        int databaseSizeBeforeUpdate = reviewRatingRepository.findAll().size();
        reviewRating.setId(count.incrementAndGet());

        // Create the ReviewRating
        ReviewRatingDTO reviewRatingDTO = reviewRatingMapper.toDto(reviewRating);

        // If url ID doesn't match entity ID, it will throw BadRequestAlertException
        restReviewRatingMockMvc
            .perform(
                put(ENTITY_API_URL_ID, count.incrementAndGet())
                    .contentType(MediaType.APPLICATION_JSON)
                    .content(TestUtil.convertObjectToJsonBytes(reviewRatingDTO))
                    .with(csrf().asHeader())
            )
            .andExpect(status().isBadRequest());

        // Validate the ReviewRating in the database
        List<ReviewRating> reviewRatingList = reviewRatingRepository.findAll();
        assertThat(reviewRatingList).hasSize(databaseSizeBeforeUpdate);
    }

    @Test
    @Timeout(value = 10, unit = TimeUnit.MINUTES)
    @Transactional
    void putWithMissingIdPathParamReviewRating() throws Exception {
        int databaseSizeBeforeUpdate = reviewRatingRepository.findAll().size();
        reviewRating.setId(count.incrementAndGet());

        // Create the ReviewRating
        ReviewRatingDTO reviewRatingDTO = reviewRatingMapper.toDto(reviewRating);

        // If url ID doesn't match entity ID, it will throw BadRequestAlertException
        restReviewRatingMockMvc
            .perform(
                put(ENTITY_API_URL)
                    .contentType(MediaType.APPLICATION_JSON)
                    .content(TestUtil.convertObjectToJsonBytes(reviewRatingDTO))
                    .with(csrf().asHeader())
            )
            .andExpect(status().isMethodNotAllowed());

        // Validate the ReviewRating in the database
        List<ReviewRating> reviewRatingList = reviewRatingRepository.findAll();
        assertThat(reviewRatingList).hasSize(databaseSizeBeforeUpdate);
    }

    @Test
    @Timeout(value = 10, unit = TimeUnit.MINUTES)
    @Transactional
    void partialUpdateReviewRatingWithPatch() throws Exception {
        // Initialize the database
        reviewRatingRepository.saveAndFlush(reviewRating);

        int databaseSizeBeforeUpdate = reviewRatingRepository.findAll().size();

        // Update the reviewRating using partial update
        ReviewRating partialUpdatedReviewRating = new ReviewRating();
        partialUpdatedReviewRating.setId(reviewRating.getId());
        partialUpdatedReviewRating.setComment(reviewRating.getComment());
        partialUpdatedReviewRating.status(UPDATED_STATUS);

        restReviewRatingMockMvc
            .perform(
                patch(ENTITY_API_URL_ID, partialUpdatedReviewRating.getId())
                    .contentType("application/merge-patch+json")
                    .content(TestUtil.convertObjectToJsonBytes(partialUpdatedReviewRating))
                    .with(csrf().asHeader())
            )
            .andExpect(status().isOk());

        // Validate the ReviewRating in the database
        List<ReviewRating> reviewRatingList = reviewRatingRepository.findAll();
        assertThat(reviewRatingList).hasSize(databaseSizeBeforeUpdate);
        ReviewRating testReviewRating = reviewRatingList.get(reviewRatingList.size() - 1);
        assertThat(testReviewRating.getComment()).isEqualTo(partialUpdatedReviewRating.getComment());
        assertThat(testReviewRating.getStatus()).isEqualTo(UPDATED_STATUS);
    }

    @Test
    @Transactional
    void fullUpdateReviewRatingWithPatch() throws Exception {
        // Initialize the database
        reviewRatingRepository.saveAndFlush(reviewRating);

        int databaseSizeBeforeUpdate = reviewRatingRepository.findAll().size();

        // Update the reviewRating using partial update
        ReviewRating partialUpdatedReviewRating = new ReviewRating();
        partialUpdatedReviewRating.setId(reviewRating.getId());
        partialUpdatedReviewRating.setComment(createUpdatedEntity(em).getComment());
        partialUpdatedReviewRating.status(UPDATED_STATUS);

        restReviewRatingMockMvc
            .perform(
                patch(ENTITY_API_URL_ID, partialUpdatedReviewRating.getId())
                    .contentType("application/merge-patch+json")
                    .content(TestUtil.convertObjectToJsonBytes(partialUpdatedReviewRating))
                    .with(csrf().asHeader())
            )
            .andExpect(status().isOk());

        // Validate the ReviewRating in the database
        List<ReviewRating> reviewRatingList = reviewRatingRepository.findAll();
        assertThat(reviewRatingList).hasSize(databaseSizeBeforeUpdate);
        ReviewRating testReviewRating = reviewRatingList.get(reviewRatingList.size() - 1);
        partialUpdatedReviewRating.getComment().setId(testReviewRating.getComment().getId());
        assertEquals(true, testReviewRating.getComment().equals(partialUpdatedReviewRating.getComment()));
        assertThat(testReviewRating.getStatus()).isEqualTo(UPDATED_STATUS);
    }

    @Test
    @Transactional
    void patchNonExistingReviewRating() throws Exception {
        int databaseSizeBeforeUpdate = reviewRatingRepository.findAll().size();
        reviewRating.setId(count.incrementAndGet());

        // Create the ReviewRating
        ReviewRatingDTO reviewRatingDTO = reviewRatingMapper.toDto(reviewRating);

        // If the entity doesn't have an ID, it will throw BadRequestAlertException
        restReviewRatingMockMvc
            .perform(
                patch(ENTITY_API_URL_ID, reviewRatingDTO.getId())
                    .contentType("application/merge-patch+json")
                    .content(TestUtil.convertObjectToJsonBytes(reviewRatingDTO))
                    .with(csrf().asHeader())
            )
            .andExpect(status().isBadRequest());

        // Validate the ReviewRating in the database
        List<ReviewRating> reviewRatingList = reviewRatingRepository.findAll();
        assertThat(reviewRatingList).hasSize(databaseSizeBeforeUpdate);
    }

    @Test
    @Timeout(value = 10, unit = TimeUnit.MINUTES)
    @Transactional
    void patchWithIdMismatchReviewRating() throws Exception {
        int databaseSizeBeforeUpdate = reviewRatingRepository.findAll().size();
        reviewRating.setId(count.incrementAndGet());

        // Create the ReviewRating
        ReviewRatingDTO reviewRatingDTO = reviewRatingMapper.toDto(reviewRating);

        // If url ID doesn't match entity ID, it will throw BadRequestAlertException
        restReviewRatingMockMvc
            .perform(
                patch(ENTITY_API_URL_ID, count.incrementAndGet())
                    .contentType("application/merge-patch+json")
                    .content(TestUtil.convertObjectToJsonBytes(reviewRatingDTO))
                    .with(csrf().asHeader())
            )
            .andExpect(status().isBadRequest());

        // Validate the ReviewRating in the database
        List<ReviewRating> reviewRatingList = reviewRatingRepository.findAll();
        assertThat(reviewRatingList).hasSize(databaseSizeBeforeUpdate);
    }

    @Test
    @Timeout(value = 10, unit = TimeUnit.MINUTES)
    @Transactional
    void patchWithMissingIdPathParamReviewRating() throws Exception {
        int databaseSizeBeforeUpdate = reviewRatingRepository.findAll().size();
        reviewRating.setId(count.incrementAndGet());

        // Create the ReviewRating
        ReviewRatingDTO reviewRatingDTO = reviewRatingMapper.toDto(reviewRating);

        // If url ID doesn't match entity ID, it will throw BadRequestAlertException
        restReviewRatingMockMvc
            .perform(
                patch(ENTITY_API_URL)
                    .contentType("application/merge-patch+json")
                    .content(TestUtil.convertObjectToJsonBytes(reviewRatingDTO))
                    .with(csrf().asHeader())
            )
            .andExpect(status().isMethodNotAllowed());

        // Validate the ReviewRating in the database
        List<ReviewRating> reviewRatingList = reviewRatingRepository.findAll();
        assertThat(reviewRatingList).hasSize(databaseSizeBeforeUpdate);
    }

    @Test
    @Transactional
    void deleteReviewRating() throws Exception {
        // Initialize the database
        reviewRatingRepository.saveAndFlush(reviewRating);

        int databaseSizeBeforeDelete = reviewRatingRepository.findAll().size();

        // Delete the reviewRating
        restReviewRatingMockMvc
            .perform(delete(ENTITY_API_URL_ID, reviewRating.getId()).accept(MediaType.APPLICATION_JSON).with(csrf().asHeader()))
            .andExpect(status().isNoContent());

        // Validate the database contains one less item
        List<ReviewRating> reviewRatingList = reviewRatingRepository.findAll();
        assertThat(reviewRatingList).hasSize(databaseSizeBeforeDelete - 1);
    }
}
